"""
Data preprocessing functions
"""

# expose the API
from ._filtering import qc_bulk_sc, qc_sc,qc_st_sc,high_variable_gene
from ._preprocessing import preprocessing
from ._rpackage import remove_batch_effect
from ._split_he_image import process_svs_image


import pytest
import os
import sys
import numpy as np
import anndata as ad
import pandas as pd
import cytobulk as ct
import scanpy as sc
from string import ascii_uppercase




# tested
@pytest.mark.skip
def test_read_adata(adata_path):
    return sc.read_h5ad(adata_path)

@pytest.mark.skip
def test_read_df(data_path):
    return pd.read_csv(data_path,index_col=0,sep='\t')


@pytest.mark.skip
def test_read_csv(data_path):
    return pd.read_csv(data_path,index_col=0,sep=',')


@pytest.mark.skip
@pytest.mark.parametrize("adata_path,bulk_path,annotation_key", [("C:/Users/wangxueying/project/CytoBulk/case/human_sc/input/filtered_A36_sample.h5ad",
                                                                    "C:/Users/wangxueying/project/CytoBulk/case/human_sc/input/A35_sample_stimulated_bulk.h5ad",
                                                                    "Manually_curated_celltype")])
def test_bulk_deconv(bulk_path,adata_path,annotation_key):
    sc = test_read_adata(adata_path)
    bulk = test_read_adata(bulk_path)
    ct.tl.bulk_deconv(bulk_data = bulk,sc_adata = sc,
                        annotation_key =annotation_key,
                        out_dir='C:/Users/wangxueying/project/CytoBulk/case/human_sc/out',dataset_name='filtered_A36_sc_35')



@pytest.mark.skip
@pytest.mark.parametrize("adata_path,st_path,annotation_key", [("C:/Users/wangxueying/project/CytoBulk/case/mouse_mob/input/sc_layer_mob.h5ad",
                                                                            "C:/Users/wangxueying/project/CytoBulk/case/mouse_mob/input/st_mob.h5ad",
                                                                            "subtype")])
def test_st_deconv(st_path,adata_path,annotation_key):
    sc = test_read_adata(adata_path)
    st = test_read_adata(st_path)
    ct.tl.st_deconv(st_adata = st,sc_adata = sc,
                        annotation_key =annotation_key,
                        out_dir='C:/Users/wangxueying/project/CytoBulk/case/MOB_layer_cut_add_7181',
                        dataset_name="MOB_layer")

@pytest.mark.skip
@pytest.mark.parametrize("adata_path,st_path,annotation_key", [("C:/Users/wangxueying/project/CytoBulk/case/PDAC/input/sc_adata.h5ad",
                                                                "C:/Users/wangxueying/project/CytoBulk/case/PDAC/input/st_adata.h5ad",
                                                                "cell_type")])
def test_st_deconv_pdac(st_path,adata_path,annotation_key):
    sc = test_read_adata(adata_path)
    st = test_read_adata(st_path)
    ct.tl.st_deconv(st_adata = st,sc_adata = sc,
                        annotation_key =annotation_key,
                        out_dir='C:/Users/wangxueying/project/CytoBulk/case/PDAC/out',
                        dataset_name="PDAC",
                        n_cell=15)
    


@pytest.mark.skip
@pytest.mark.parametrize("frac_data,sc_adata,bulk_adata,n_cell,annotation_key", [("../data/output/prediction_frac.csv",
                                                                            "../data/filtered/sc_data_filtered_A36_sc_35.h5ad",
                                                                            "../data/filtered/bulk_data_filtered_A36_sc_35.h5ad",
                                                                            100,
                                                                            "curated_cell_type")])
def test_bulk_mapping(frac_data,sc_adata,bulk_adata,n_cell,annotation_key):
    frac_data = test_read_csv(frac_data)
    sc_adata = test_read_adata(sc_adata)
    bulk_adata = test_read_adata(bulk_adata)
    ct.tl.bulk_mapping(frac_data = frac_data,
                        sc_adata = sc_adata,
                        bulk_adata = bulk_adata,
                        n_cell = n_cell,
                        annotation_key=annotation_key,
                        dataset_name="filtered_A36_sc_35",
                        out_dir="../data")


@pytest.mark.skip
@pytest.mark.parametrize("adata_path,bulk_path,marker_path,annotation_key,out_dir,dataset_name", [("../data/input_data/filtered_A36_sample_sc.h5ad",
                                                                            "../data/input_data/filtered_A35_sample_st.h5ad",
                                                                            "../data/input_data/cell_meta.xlsx",
                                                                            "Manually_curated_celltype",
                                                                            './human_sc_st',
                                                                            'human_st')])
def test_st_deconv_human(bulk_path,adata_path,marker_path,annotation_key,out_dir,dataset_name):
    sc = test_read_adata(adata_path)
    st = test_read_adata(bulk_path)
    marker = pd.read_excel(marker_path, sheet_name = "marker")
    names = pd.read_excel(marker_path, sheet_name = "rename")
    names = names.set_index(['Original_name'])['Curated_name'].to_dict()
    ct.tl.st_deconv(st_adata = st,sc_adata = sc,marker_data=marker,
                        annotation_key =annotation_key,
                        rename = names,
                        out_dir=out_dir,
                        dataset_name=dataset_name)
    


@pytest.mark.skip
@pytest.mark.parametrize("adata_path,bulk_path,marker_path,annotation_key,out_dir,dataset_name", [("D:/project/CytoBulk/CytoBulk/tests/ST_HNSC_GSE139324/input_data/HNSC_GSE139324.h5ad",
                                                                            "D:/project/CytoBulk/CytoBulk/tests/ST_HNSC_GSE139324/input_data/HNSC_GSE139324_expression_test.csv",
                                                                            "D:/project/CytoBulk/CytoBulk/tests/ST_HNSC_GSE139324/input_data/cell_meta.xlsx",
                                                                            "Celltype..minor.lineage.",
                                                                            './ST_HNSC_GSE139324',
                                                                            'HNSC_GSE139324')])
def test_st_deconv_hnsc(bulk_path,adata_path,marker_path,annotation_key,out_dir,dataset_name):
    sc_adata = test_read_adata(adata_path)
    sc_adata.__dict__['_raw'].__dict__['_var'] = sc_adata.__dict__['_raw'].__dict__['_var'].rename(columns={'_index': 'features'})
    st = pd.read_csv(bulk_path,index_col=0)
    st_adata = sc.AnnData(st)
    marker = pd.read_excel(marker_path, sheet_name = "marker")
    names = pd.read_excel(marker_path, sheet_name = "rename")
    names = names.set_index(['Original_name'])['Curated_name'].to_dict()
    ct.tl.st_deconv(st_adata = st_adata,sc_adata = sc_adata,marker_data=marker,
                        annotation_key =annotation_key,
                        rename = names,
                        out_dir=out_dir,
                        dataset_name=dataset_name)




@pytest.mark.skip
@pytest.mark.parametrize("sc_adata,marker_path,annotation_key,out_dir,dataset_name", [("../data/input_data/MM_GSE151310.h5ad",
                                                                            "../data/input_data/cell_meta.xlsx",
                                                                            "Celltype..minor.lineage.",
                                                                            './MM_25',
                                                                            'MM')])
def test_simulation_st(sc_adata,dataset_name, out_dir, annotation_key,marker_path):
    sc = test_read_adata(sc_adata)
    marker = pd.read_excel(marker_path, sheet_name = "marker")
    names = pd.read_excel(marker_path, sheet_name = "rename")
    names = names.set_index(['Original_name'])['Curated_name'].to_dict()
    sc.__dict__['_raw'].__dict__['_var'] = sc.__dict__['_raw'].__dict__['_var'].rename(columns={'_index': 'features'})
    common_cell = names.keys()
    pseudo_st = ct.ul.st_simulation_case(
                sc, 
                common_cell, 
                annotation_key = annotation_key,
                project=dataset_name, 
                out_dir=out_dir,
                n_sample_each_group=100,
                min_cells_each_group=25,
                cell_gap_each_group=1,
                group_number=5,
                rename_dict=names,
                return_adata=True,
                save=True)

@pytest.mark.skip
@pytest.mark.parametrize("sc_adata,st_adata,marker_path,annotation_key,out_dir,dataset_name", [("../data/input_data/MM_GSE151310.h5ad",
                                                                            "../data/input_data/stimulated_MM_st_5.h5ad",
                                                                            "../data/input_data/cell_meta.xlsx",
                                                                            "Celltype..minor.lineage.",
                                                                            './MM_5_bulk',
                                                                            'MM_5_bulk')])   
def test_st_deconv_new(sc_adata,st_adata,marker_path,annotation_key,out_dir,dataset_name):
    sc = test_read_adata(sc_adata)
    st = test_read_adata(st_adata)
    sc.__dict__['_raw'].__dict__['_var'] = sc.__dict__['_raw'].__dict__['_var'].rename(columns={'_index': 'features'})
    marker = pd.read_excel(marker_path, sheet_name = "marker")
    names = pd.read_excel(marker_path, sheet_name = "rename")
    names = names.set_index(['Original_name'])['Curated_name'].to_dict()
    ct.tl.st_deconv(st_adata = st,sc_adata = sc,marker_data=marker,
                        annotation_key =annotation_key,
                        rename = names,
                        out_dir=out_dir,
                        dataset_name=dataset_name,
                        different_source=True,
                        n_cell=5)


@pytest.mark.parametrize("st_path", [("C:/Users/wangxueying/project/CytoBulk/case/ER2/ST_CID4535_374.h5ad")])   
def test_segmentation(st_path):
    st = test_read_adata(st_path)
    ret,cell_pos = ct.tl.predict_cell_num(
                    st,
                    out_dir="C:/Users/wangxueying/project/CytoBulk/case/ER2",
                    diameter=None,
                    save_png_result=False,
                    model_type='cyto3',
                    cellprob_threshold=0.4
                )



@pytest.mark.skip
@pytest.mark.parametrize("sc_adata,st_adata", [("/Users/wangxueying/project/cytobulk/data/sc_adata.h5ad",
                                                "/Users/wangxueying/project/cytobulk/data/st_adata_sub_4_construction.h5ad")])   
def test_st_mapping(sc_adata,st_adata):
    sc = test_read_adata(sc_adata)
    st = test_read_adata(st_adata)
    ct.tl.st_mapping(st_adata = st,sc_adata = sc)

if __name__ == '__main__':
    pytest.main(["-s", "test_tl_tools.py"])


import pytest
import os
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import numpy as np
import anndata as ad
import pandas as pd
import cytobulk as ct
import scanpy as sc
from string import ascii_uppercase


@pytest.mark.skip
def test_read_adata(adata_path):
    return sc.read_h5ad(adata_path)


@pytest.mark.parametrize("adata_path, project, out_dir,annotation_key,marker_path", [("../data/A35_sample.h5ad","A35_sample","../data","Manually_curated_celltype",
                                                                                    "../data/cell_meta.xlsx")])

@pytest.mark.skip
def test_stimulation_bulk(adata_path,project, out_dir, annotation_key,marker_path):
    sc = test_read_adata(adata_path)
    marker = pd.read_excel(marker_path, sheet_name = "marker")
    names = pd.read_excel(marker_path, sheet_name = "rename")
    names = names.set_index(['Original_name'])['Curated_name'].to_dict()
    common_cell = names.keys()
    pseudo_bulk, pseudo_prop = ct.ul.bulk_simulation(sc, 
                                common_cell, 
                                annotation_key = annotation_key,
                                project=project, 
                                out_dir=out_dir,
                                n_sample_each_group=100,
                                min_cells_each_group=100,
                                cell_gap_each_group=100,
                                group_number=5,
                                rename_dict=names,
                                save=True)
    assert isinstance(pseudo_bulk,pd.DataFrame)
    assert isinstance(pseudo_prop,pd.DataFrame)
    assert len(common_cell)==len(pseudo_prop.columns)

@pytest.mark.parametrize("adata_path, project, out_dir,annotation_key,marker_path", [("../data/A35_sample.h5ad","A35_sample","../data","Manually_curated_celltype",
                                                                                    "../data/cell_meta.xlsx")])
def test_stimulation_case(adata_path,project, out_dir, annotation_key,marker_path):
    sc = test_read_adata(adata_path)
    marker = pd.read_excel(marker_path, sheet_name = "marker")
    names = pd.read_excel(marker_path, sheet_name = "rename")
    names = names.set_index(['Original_name'])['Curated_name'].to_dict()
    common_cell = names.keys()
    pseudo_bulk, pseudo_prop = ct.ul.bulk_simulation_case(
                                sc, 
                                common_cell, 
                                annotation_key = annotation_key,
                                project=project, 
                                out_dir=out_dir,
                                n_sample_each_group=100,
                                min_cells_each_group=100,
                                cell_gap_each_group=100,
                                group_number=5,
                                rename_dict=names,
                                save=True)
    assert isinstance(pseudo_bulk,pd.DataFrame)
    assert isinstance(pseudo_prop,pd.DataFrame)
    assert len(common_cell)==len(pseudo_prop.columns)



if __name__ == '__main__':
    pytest.main(["-s", "test_ul_stimulation.py"])


import pytest
import os
import sys
import numpy as np
import anndata as ad
import pandas as pd
import cytobulk as ct
import scanpy as sc
from string import ascii_uppercase

# tested
@pytest.mark.skip
def test_read_adata(adata_path):
    return sc.read_h5ad(adata_path)

@pytest.mark.skip
def test_read_df(data_path):
    return pd.read_csv(data_path,index_col=0,sep='\t')



@pytest.mark.skip
@pytest.mark.parametrize("sc_adata,st_adata", [("C:/Users/wangxueying/project/CytoBulk/case/mouse_mob/input/sc_layer_mob.h5ad",
                                                "C:/Users/wangxueying/project/CytoBulk/case/MOB_layer_cut_add_7181/output/MOB_layer_st_adata.h5ad")])   
def test_st_mapping_mouse(sc_adata,st_adata):
    sc = test_read_adata(sc_adata)
    st = test_read_adata(st_adata)
    ct.tl.st_mapping(st_adata = st,sc_adata = sc,
                     out_dir="C:/Users/wangxueying/project/CytoBulk/case/MOB_layer_cut_add_7181",
                     project="mouse_mob",
                     annotation_key='subtype',
                     mean_cell_numbers=18)


@pytest.mark.skip
@pytest.mark.parametrize("sc_adata,st_adata", [("C:/Users/wangxueying/project/CytoBulk/case/PDAC/input/sc_adata.h5ad",
                                                "C:/Users/wangxueying/project/CytoBulk/case/PDAC/out/output/PDAC_st_adata.h5ad")])   
def test_st_mapping_pdac(sc_adata,st_adata):
    sc = test_read_adata(sc_adata)
    st = test_read_adata(st_adata)
    ct.tl.st_mapping(st_adata = st,sc_adata = sc,
                     out_dir="C:/Users/wangxueying/project/CytoBulk/case/PDAC/out",
                     project="pdac",
                     annotation_key='cell_type',
                     mean_cell_numbers=18)
    

@pytest.mark.skip
@pytest.mark.parametrize("sc_adata,st_adata", [("C:/Users/wangxueying/project/CytoBulk/case/10x/input/sc_adata.h5ad",
                                                "C:/Users/wangxueying/project/CytoBulk/case/10x/sub4/st_adata_sub_4_construction.h5ad")])   
def test_st_mapping(sc_adata,st_adata):
    sc = test_read_adata(sc_adata)
    st = test_read_adata(st_adata)
    ct.tl.st_mapping(st_adata = st,sc_adata = sc,
                     out_dir="C:/Users/wangxueying/project/CytoBulk/case/10x/sub4",
                     project="sub4",
                     annotation_key='cell_type',
                     mean_cell_numbers=15)
@pytest.mark.skip   
@pytest.mark.parametrize("sc_adata,st_adata", [("C:/Users/wangxueying/project/CytoBulk/case/10x/input/sc_adata.h5ad",
                                                "C:/Users/wangxueying/project/CytoBulk/case/10x/sub6/st_adata_sub_6_deconv.h5ad")])   
def test_st_mapping_sub6(sc_adata,st_adata):
    sc = test_read_adata(sc_adata)
    st = test_read_adata(st_adata)
    ct.tl.st_mapping(st_adata = st,sc_adata = sc,
                     out_dir="C:/Users/wangxueying/project/CytoBulk/case/10x/sub6",
                     project="sub6",
                     annotation_key='cell_type')
@pytest.mark.skip 
@pytest.mark.parametrize("sc_adata,st_adata", [("C:/Users/wangxueying/project/CytoBulk/case/10x/input/sc_adata.h5ad",
                                                r"C:\Users\wangxueying\project\CytoBulk\case\10x\sub4\st_adata_sub_4_construction.h5ad")])   
def test_st_mapping_sub4(sc_adata,st_adata):
    sc = test_read_adata(sc_adata)
    st = test_read_adata(st_adata)
    print(st.uns.keys())
    ct.tl.st_mapping(st_adata = st,sc_adata = sc,
                     out_dir="C:/Users/wangxueying/project/CytoBulk/case/10x/sub4",
                     project="sub4",
                     annotation_key='cell_type')
@pytest.mark.skip      
@pytest.mark.parametrize("sc_adata,st_adata", [("C:/Users/wangxueying/project/CytoBulk/case/10x/input/sc_adata.h5ad",
                                                "C:/Users/wangxueying/project/CytoBulk/case/10x/sub5/st_adata_sub_5_deconv.h5ad")])   
def test_st_mapping_sub5(sc_adata,st_adata):
    sc = test_read_adata(sc_adata)
    st = test_read_adata(st_adata)
    ct.tl.st_mapping(st_adata = st,sc_adata = sc,
                     out_dir="C:/Users/wangxueying/project/CytoBulk/case/10x/sub5",
                     project="sub5",
                     annotation_key='cell_type')
    
         
@pytest.mark.parametrize("sc_adata,bulk_adata", [(r"C:\Users\wangxueying\project\CytoBulk\case\he_image\svs\TCGA_LUSC\sub_HTAN_MSK.h5ad",
                                                r"C:\Users\wangxueying\project\CytoBulk\case\he_image\svs\TCGA_LUSC\lusc_bulk_adata.h5ad")])   
def test_bulk_mapping_sub(sc_adata,bulk_adata):
    sc = test_read_adata(sc_adata)
    bulk = test_read_adata(bulk_adata)
    ct.tl.bulk_mapping(bulk_adata = bulk,
                       sc_adata = sc,
                        out_dir=r"C:\Users\wangxueying\project\CytoBulk\case\he_image\svs\TCGA_LUSC",
                        project="test",
                        n_cell=1000,
                        cpu_num=2,
                        annotation_key='he_cell_type')



import pytest
import os
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import numpy as np
import anndata as ad
import pandas as pd
import cytobulk as ct
import scanpy as sc
from string import ascii_uppercase

# tested
@pytest.mark.skip
def test_read_adata(adata_path):
    return sc.read_h5ad(adata_path)

@pytest.mark.skip
def test_read_df(data_path):
    return pd.read_csv(data_path,index_col=0,sep='\t')

# tested
@pytest.mark.skip
@pytest.mark.parametrize("adata_path", [("../data/A36_sample.h5ad")])
def test_sc_qc(adata_path):
    sc = test_read_adata(adata_path)
    assert isinstance(ct.pp.qc_sc(sc,save=True,out_dir='../data',project='A36_sample'),ad.AnnData)
    #return ct.pp.qc_sc(sc)   
@pytest.mark.skip
@pytest.mark.parametrize("adata_path,bulk_path", [("../data/A36_sample.h5ad","../data/reference_bulk_data/A35_sample_stimulated_bulk.txt")])
def test_qc_bulk_sc(bulk_path,adata_path):
    sc = test_read_adata(adata_path)
    bulk = test_read_df(bulk_path)
    ct.pp.qc_bulk_sc(bulk_data = bulk,sc_adata = sc,save=True,out_dir='../data',dataset_name='A36_sc_35_bulk')



@pytest.mark.parametrize("adata_path,bulk_path,marker_path,annotation_key", [("../data/filtered_A36_sample.h5ad",
                                                                            "../data/reference_bulk_data/A35_sample_stimulated_bulk.txt",
                                                                            "../data/cell_meta.xlsx",
                                                                            "Manually_curated_celltype")])
def test_preprocessing_bulk_sc(bulk_path,adata_path,marker_path,annotation_key):
    sc = test_read_adata(adata_path)
    bulk = test_read_df(bulk_path)
    marker = pd.read_excel(marker_path, sheet_name = "marker")
    names = pd.read_excel(marker_path, sheet_name = "rename")
    names = names.set_index(['Original_name'])['Curated_name'].to_dict()
    ct.pp.preprocessing(bulk_data = bulk,sc_adata = sc,marker_data=marker,
                        annotation_key =annotation_key,
                        rename = names,
                        out_dir='../data',dataset_name='filtered_A36_sc_35')
    


if __name__ == '__main__':
    pytest.main(["-s", "test_pl_preprocessing.py"])



import pandas as pd
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib.colors import to_hex
import matplotlib.patches
import matplotlib.cm as cm
import scanpy as sc
from .. import utils
from .. import get

class Const:
    """
    Some COLOR SET used in the class.
    
    """
    DIFF_COLOR_2 = ['#011F5B','#F19CBB']
    #DIFF_COLOR_N = ['#005f73','#94d2bd','#e9d8a6','#fac748','#e76f51','#9b2226']
    DIFF_COLOR_N = ['#191970','#ADD8E6','#005f73','#94d2bd','#e9d8a6','#fac748','#e76f51','#9b2226']




def _plot_scatter_2label(x1,x2,X1_label,X2_label,title,color_set=None,fig_format=None,out_dir='/out'):
    columns=x1.columns
    plt.figure(figsize=(4, 4))
    if color_set is None:
        color_set = Const.DIFF_COLOR_2
    if fig_format is None:
        fig_format = 'svg'
    s = [0,1]
    data = [x1,x2]
    marker1 = ["^", "o"]
    for index in range(2):
        XOffset = data[index][columns[0]]
        YOffset = data[index][columns[1]]
        s[index] = plt.scatter(XOffset, YOffset, c=color_set[index], s=50, alpha=1, marker=marker1[index], linewidth=0,zorder=-index) 

    plt.legend((s[0],s[1]),(X1_label,X2_label) ,loc = 'best')
    plt.xlabel("X1")
    plt.ylabel("X2")
    plt.title(f'Scatter Plot of {title}',fontsize=10) 
    plt.savefig(f'{out_dir}/scatter_{title}.{fig_format}', bbox_inches='tight',transparent=True)


def plot_batch_effect(bulk_adata,pseudo_bulk,out_dir,title=''):
    def _get_paired_data(adata,status=None):
        data = get.count_data(adata,counts_location=status)
        return(utils.pca(data))
    def _plot_paired_data(bulk_adata,pseudo_bulk,title,out_dir,status=None):
        X1 = _get_paired_data(bulk_adata,status)
        X2 = _get_paired_data(pseudo_bulk,status)
        _plot_scatter_2label(X1,X2,X1_label="input_data",X2_label="simulated data",title=title,out_dir=out_dir)

    _plot_paired_data(bulk_adata,pseudo_bulk,title+"(original)",out_dir=out_dir)
    _plot_paired_data(bulk_adata,pseudo_bulk,title+"(batch effected)",out_dir=out_dir,status="batch_effected")        

def plot_celltype_fraction_pie(adata,
                               scale_facter_x=1,
                               scale_factor_y=1,
                               r=0.1,
                               out_dir='/plot',
                               project_name='project',
                               color_list=None,
                               color_map="Spectral",
                               rotation_angle=None,
                               figsize=(2, 2)):
    """
    Plot cell type fraction as pie charts at spatial coordinates.

    This function generates pie charts representing the cell type fraction at each spatial location (from the 'spatial' data in `adata`) and plots them on a 2D scatter plot. Each pie chart is positioned based on the spatial coordinates and reflects the relative abundance of each cell type at that spot. The resulting figure is saved as an SVG file.

    Parameters
    ----------
    adata : anndata.AnnData
        An :class:`~anndata.AnnData` object containing spatial transcriptomics data.
        The spatial coordinates should be stored in `adata.obsm['spatial']`, and the deconvolution results (cell type fractions) should be stored in `adata.uns['deconv']`.
    
    scale_facter_x : float, optional (default: 1)
        A scaling factor to apply to the X-axis coordinates of the spatial data.
    
    scale_factor_y : float, optional (default: 1)
        A scaling factor to apply to the Y-axis coordinates of the spatial data.
    
    r : float, optional (default: 0.1)
        Radius of the pie charts representing cell type fractions at each spatial coordinate.
    
    out_dir : string, optional (default: '/plot')
        The directory where the output figure (SVG format) will be saved.
    
    project_name : string, optional (default: 'project')
        The prefix to use for the saved SVG file.
    
    color_list : list, optional (default: None)
        A list of colors to use for the cell types. If `None`, a colormap (specified by `color_map`) will be used to generate colors for each cell type.
    
    color_map : string, optional (default: 'Spectral')
        The name of the colormap to use if `color_list` is not provided. It will generate a set of colors for the different cell types.
    
    rotation_angle : float, optional (default: None)
        An optional angle (in degrees) to rotate the spatial coordinates. If provided, the rotation is applied to the coordinates before plotting.
    
    figsize : tuple, optional (default: (2, 2))
        The size of the output figure, specified as a tuple of (width, height) in inches.
    
    Returns
    -------
    None
        The function does not return any values but saves the generated pie chart plot as an SVG file in the specified `out_dir`.
    """
    loc_xy=adata.obsm['spatial']
    loc_xy=pd.DataFrame(loc_xy,columns=['x','y'],index=adata.obs_names)
    if rotation_angle:
        new_loc = loc_xy.apply(lambda x : utils.rotate_matrix(x['x'],x['y'],rotation_angle), axis=1)
        loc_xy = pd.DataFrame(list(new_loc),columns=loc_xy.columns,index=loc_xy.index)
    cell_type_fraction = adata.uns['deconv']
    cell_type_fraction['None']=0
    cell_type_fraction.loc[(cell_type_fraction==0).all(axis=1),"None"]=1
    cell_type_fraction['x']=loc_xy['x']
    cell_type_fraction['y']=loc_xy['y']
    loc=cell_type_fraction
    loc['max_idx'] = loc.iloc[:,:-2].idxmax(axis=1)
    loc['x']=loc['x']/scale_facter_x
    loc['y']=loc['y']/scale_factor_y
    fig, ax = plt.subplots(figsize=figsize)
    def _plot_pie(x, ax, color_sets,r): 
        fraction = x[:-3]
        ax.pie(fraction, center=(x['x'],x['y']), radius=r,colors=[color_sets[key] for key in loc.columns[:-3]])
        ax.set_title('CytoBulk',fontsize=20,x=1.2,y=1.7)
    # git min/max values for the axes
    cells = loc.columns[:-3].tolist()
    if color_list:
        color_dict = {item: color for item, color in zip(cells, color_list)}
    else:
        color_map = cm.get_cmap(color_map)
        list_length = len(cells)
        colors_list = [to_hex(color_map(i / (list_length - 1))) for i in range(list_length)]
        color_dict = {item: color for item, color in zip(cells, colors_list)}
    
    loc.apply(lambda x : _plot_pie(x, ax,color_dict,r=0.1), axis=1)
    
    handles = []
    for i, l in enumerate(cells):
        handles.append(matplotlib.patches.Patch(color=colors_list[i], label=cells[i]))
    plt.legend(handles,cells, bbox_to_anchor=(0.1,0.1),fontsize=12,markerscale=0.5,ncol=3,loc="upper left")
    fig.show()
    fig.savefig(f"{out_dir}/{project_name}_cell_fraction_pie.svg", format="svg", dpi=1200,bbox_inches = 'tight',transparent = True)


def plot_celltype_fraction_heatmap(adata,
                                   label,
                                    r=0.1,
                                    out_dir='/plot',
                                    project_name='project',
                                    color_map='crest',
                                    rotation_angle=None,
                                    figsize=(2.7, 2)):
    """
    Plot a heatmap of cell type fractions at spatial coordinates.

    This function visualizes the fraction of a specified cell type as a heatmap on a 2D scatter plot, using spatial transcriptomics data. The color intensity of each point corresponds to the cell type fraction at that spatial location. The resulting figure is saved as an SVG file.

    Parameters
    ----------
    adata : anndata.AnnData
        An :class:`~anndata.AnnData` object containing spatial transcriptomics data.
        The spatial coordinates should be stored in `adata.obsm['spatial']`, and the deconvolution results (cell type fractions) should be stored in `adata.uns['deconv']`.
    
    label : string
        The name of the cell type to plot. This should correspond to one of the columns in `adata.uns['deconv']`, which represents the fractions of different cell types.
    
    r : float, optional (default: 0.1)
        Radius of the points representing the cell type fractions at each spatial coordinate.
    
    out_dir : string, optional (default: '/plot')
        The directory where the output figure (SVG format) will be saved.
    
    project_name : string, optional (default: 'project')
        The prefix to use for the saved SVG file.
    
    color_map : string, optional (default: 'crest')
        The name of the colormap to use for the heatmap. This can be any colormap recognized by `seaborn` or `matplotlib`.
    
    rotation_angle : float, optional (default: None)
        An optional angle (in degrees) to rotate the spatial coordinates. If provided, the rotation is applied to the coordinates before plotting.
    
    figsize : tuple, optional (default: (2.7, 2))
        The size of the output figure, specified as a tuple of (width, height) in inches.
    
    Returns
    -------
    None
        The function does not return any values but saves the generated heatmap plot as an SVG file in the specified `out_dir`.
    """
    from sklearn.preprocessing import minmax_scale
    fig, ax = plt.subplots(figsize=figsize)
    plt.xticks([])
    plt.yticks([])
    plt.axis('off')
    loc_xy=adata.obsm['spatial']
    loc_xy=pd.DataFrame(loc_xy,columns=['x','y'],index=adata.obs_names)
    if rotation_angle:
        new_loc = loc_xy.apply(lambda x : utils.rotate_matrix(x['x'],x['y'],rotation_angle), axis=1)
        loc_xy = pd.DataFrame(list(new_loc),columns=loc_xy.columns,index=loc_xy.index)
    cell_type_fraction = adata.uns['deconv']
    cell_type_fraction['x']=loc_xy['x']
    cell_type_fraction['y']=loc_xy['y']
    cell_type_fraction[label] = minmax_scale(cell_type_fraction[label].values)
    sns.scatterplot(data=cell_type_fraction, x="x", y="y",palette=color_map,hue=label,s=20,legend=False,edgecolor="None",markers="pentagon")
    norm = plt.Normalize(cell_type_fraction[label].min(), cell_type_fraction[label].max())
    cmap = sns.color_palette(color_map, as_cmap=True)
    sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
    sm.set_array([])
    cax = fig.add_axes([ax.get_position().x1+0.05, ax.get_position().y0, 0.03, ax.get_position().height])
    ax.set_title(f'{label}',fontsize=10,x=ax.get_position().x1/2,y=ax.get_position().y1+0.1)
    ax.figure.colorbar(sm, cax=cax)
    fig.show()
    fig.savefig(f"{out_dir}/{project_name}_{label}.svg", format="svg", dpi=1200,bbox_inches = 'tight',transparent = True)

    
def plot_paired_violin(adata,
                       label,
                       gene,
                       stats_method='spearmanr',
                       out_dir='/plot',
                       color_list=Const.DIFF_COLOR_N,
                       project_name="test",
                       figsize=(6, 4),
                       ylim=[-0.1, 1.2]):
    """
    Plot paired violin plots for gene expression and predicted cell type fractions.

    This function generates paired violin plots comparing the expression of a specific gene with the predicted cell type fraction (from deconvolution data) for the same set of observations. It also computes a statistical correlation (Spearman or Pearson) between the gene expression and the predicted cell type fractions and displays the correlation coefficient and significance on the plot.

    Parameters
    ----------
    adata : anndata.AnnData
        An :class:`~anndata.AnnData` object containing gene expression data and the output of a cell type deconvolution.
        - Gene expression data should be stored in `adata.X` with cell/barcode names in `adata.obs_names` and gene names in `adata.var_names`.
        - Predicted cell type fractions (from deconvolution) should be stored in `adata.uns['deconv']`.
    
    label : string
        The name of the cell type (as stored in `adata.uns['deconv']`) to be compared against the gene expression.
    
    gene : string
        The name of the gene (as stored in `adata.var_names`) whose expression levels will be compared to the predicted cell type fractions.
    
    stats_method : string, optional (default: 'spearmanr')
        The statistical method used to compute the correlation between gene expression and predicted cell type fractions.
        - `'spearmanr'`: Spearman’s rank correlation.
        - `'pearsonr'`: Pearson’s correlation.
    
    out_dir : string, optional (default: '/plot')
        The directory where the output violin plot (SVG format) will be saved.
    
    color_list : list, optional (default: Const.DIFF_COLOR_N)
        A list of colors to use for the violin plots. If not provided, a default color palette will be used.
    
    project_name : string, optional (default: 'test')
        The prefix to use for the saved SVG file.
    
    figsize : tuple, optional (default: (6, 4))
        The size of the output figure, specified as a tuple of (width, height) in inches.
    
    ylim : list, optional (default: [-0.1, 1.2])
        The limits for the Y-axis of the violin plot. This should be a list of two values [min, max].
    
    Returns
    -------
    None
        The function does not return any values but saves the generated violin plot as an SVG file in the specified `out_dir`.
    """
    from scipy import stats
    from sklearn.preprocessing import minmax_scale
    fig = plt.figure(figsize=figsize)
    color_panel = sns.set_palette(color_list)
    name_list=[]
    value_list=[]
    exp_data = pd.DataFrame(adata.X,index=adata.obs_names,columns=adata.var_names)
    predicted = adata.uns['deconv']
    common_label=exp_data.index.intersection(predicted.index)
    predicted=predicted.loc[common_label,:]
    exp_data=exp_data.loc[common_label,:]
    exp_data[gene] = minmax_scale((exp_data[gene]), feature_range=(0, 1))
    predicted[label] = minmax_scale(predicted[label], feature_range=(0, 1))
    if stats_method=='spearmanr':
        stat,p_value = stats.spearmanr(exp_data[gene],predicted[label])
    else:
        stat,p_value = stats.pearsonr(exp_data[gene],predicted[label])
    p_value=utils.convert_pvalue_to_asterisks(p_value)
    value_list.extend(exp_data[gene].values)
    name_list.extend([gene]*len(exp_data[gene].values))
    value_list.extend(predicted[label])
    name_list.extend([label]*len(predicted[label]))
    input_data=pd.DataFrame.from_dict({'source':name_list,'value':value_list})
    ax = sns.violinplot(x="source",y="value",data=input_data,hue='source',palette=color_panel)
    x1, x2 = 0, 1
    y,h = predicted[label].max()+.1,.1
    ax.plot([x1, x1, x2, x2], [y, y+h, y+h, y], lw=1, c="k") 
    ax.text((x1+x2)*.5, y+h, stats_method+": "+ str("%.2f" % stat) +" "+  p_value, ha='center', va='bottom', color="k",fontsize=14)
    ax.set_ylim([-0.3, 1.5])
    ax.tick_params(which='major',direction='in',length=3,width=1.,labelsize=14,bottom=False)
    plt.savefig(f"{out_dir}/{project_name}_{label}_{gene}_violin.svg", format="svg", dpi=1200,bbox_inches = 'tight',transparent = True)



def plot_reconstruction(adata,
                        out_dir,
                        project_name="test",
                        rotation_angle=None,
                        spot_size=0.5):
    """
    Plot reconstructed spatial transcriptomics data with correlation analysis.

    This function computes the Pearson correlation between the original and reconstructed spatial transcriptomics data for each spot. It then visualizes the spatial distribution of correlation coefficients using a scatter plot.

    Parameters
    ----------
    adata : anndata.AnnData
        An :class:`~anndata.AnnData` object containing the spatial transcriptomics data.
        - Original data should be in `adata.layers['original_st']`.
        - Reconstructed data should be in `adata.X`.
        - Spatial coordinates should be in `adata.obsm['spatial']`.

    out_dir : string
        The directory where the output plot (SVG format) will be saved.

    project_name : string, optional (default: 'test')
        The prefix to use for the saved SVG file.

    rotation_angle : float, optional
        The angle to rotate the spatial coordinates, if any.

    spot_size : float, optional (default: 0.5)
        The size of the spots in the scatter plot.

    Returns
    -------
    None
        The function does not return any values but saves the generated scatter plot as an SVG file in the specified `out_dir`.
    """
    from scipy.stats import pearsonr
    adata.var_names_make_unique()
    # Calculate Pearson correlation and p-value for each sample (row)
    loc_xy=adata.obsm['spatial']
    loc_xy=pd.DataFrame(loc_xy,columns=['x','y'],index=adata.obs_names)
    if rotation_angle:
        new_loc = loc_xy.apply(lambda x : utils.rotate_matrix(x['x'],x['y'],rotation_angle), axis=1)
        loc_xy = pd.DataFrame(list(new_loc),columns=loc_xy.columns,index=loc_xy.index)

    results = []
    for i in range(adata.shape[0]):  # Loop through each sample
        r, p = pearsonr(adata.layers['original_st'][i], adata.X[i])
        results.append((i, r, p))  # Store index, Pearson r, and p-value

    # Convert results to a DataFrame
    results_df = pd.DataFrame(results, columns=['spot', 'Pearson R', 'pvalue'])
    results_df.set_index('spot',inplace=True)
    adata.obsm['spatial']=loc_xy.values
    adata.obs = results_df
    average_corr = round(np.mean(results_df['Pearson R']),3)

    fig=sc.pl.spatial(
        adata,
        color='Pearson R',
        img_key=None,
        alpha=0.8,
        color_map="mako_r",
        size=1.5,
        title=f'Reconstructed ST\nmean Pearson correlation = {average_corr}\ngene number = {len(adata.var_names)}',
        frameon=False,
        spot_size=spot_size,
        outline_width=0,
        return_fig=True
    )
    plt.savefig(f"{out_dir}/{project_name}_reconstructed_correlation.svg", format="svg", dpi=1200,bbox_inches = 'tight',transparent = True)


    



    
def plot_he_cell_type(data,out_dir):
    """
    Visualize and save a scatter plot of cell locations by cell type.

    This function generates a scatter plot to visualize the spatial distribution of different cell types 
    based on their x and y coordinates. The plot is customized with specific colors for predefined cell types 
    and is saved as a PNG file.

    Parameters
    ----------
    data : pandas.DataFrame
        A pandas DataFrame containing the following columns:
        - 'cell_type': Categorical data representing cell types (e.g., "Epithelial Cells").
        - 'x': Numerical values representing the x-coordinates of the cells.
        - 'y': Numerical values representing the y-coordinates of the cells.

    out_dir : str
        The directory where the generated plot image will be saved. The plot will be saved as 
        "cell_type.png" within this directory.

    Returns
    -------
    None
        This function does not return any values. It generates and saves the scatter plot as a PNG file.

    Notes
    -----
    - The plot uses predefined colors for the following cell types:
        * "Epithelial Cells": #A52A2A (brown)
        * "Neutrophils": #0000B8 (blue)
        * "Plasma Cells": #0D98BA (cyan)
        * "Connective Tissue": #FFCC33 (yellow)
        * "Lymphocytes": #B284BE (purple)
      Cell types not listed above are assigned the color "black".
    - The y-axis is inverted to match the typical orientation of spatial data.
    """
    import pandas as pd
    import matplotlib.pyplot as plt

    categories = data["cell_type"].unique()
    colors = {
        "Epithelial Cells": "#A52A2A",  
        "Neutrophils": "#0000B8",      
        "Plasma Cells": "#0D98BA",     
        "Connective Tissue": "#FFCC33", 
        "Lymphocytes": "#B284BE"      
    }

    plt.figure(figsize=(4, 4))

    for category in categories:
        subset = data[data["cell_type"] == category]
        plt.scatter(subset["x"], subset["y"], label=category, color=colors.get(category, "black"), s=10, alpha=0.8)
    plt.legend(
        loc="upper right",             # Place the legend in the top-right corner
        title="Cell Types",            # Add a title to the legend
        fontsize=8,                    # Set font size for the legend labels
        title_fontsize=10,             # Set font size for the legend title
        markerscale=1.5                # Scale the legend markers (scatter points)
    )
    
    
    plt.gca().axis("off")
    plt.gca().invert_yaxis()
    save_path = f"{out_dir}/cell_type.png"
    plt.tight_layout()
    plt.savefig(save_path, dpi=300, bbox_inches="tight", pad_inches=0) 
    plt.show()


def plot_gene_similarity(adata, marker_df, custom_palette="Spectral"):
    """
    Compute and visualize cosine similarity for marker genes across cell types.

    This function calculates cosine similarity for marker genes between the gene expression matrix 
    and a mapping layer in an AnnData object. It generates a swarm plot to visualize the results 
    grouped by cell type.

    Parameters
    ----------
    adata : anndata.AnnData
        An :class:`~anndata.AnnData` object containing gene expression data in `.X` and a mapping layer in `.layers["mapping_ori"]`.

    marker_df : pandas.DataFrame
        DataFrame containing marker gene information. It must include the following columns:
        - 'gene_symbol': Gene symbol names.
        - 'cell_type': Cell type associated with the marker gene.
        - 'pvalue': P-value indicating marker significance.
        - 'score': Score representing the marker strength.

    custom_palette : str or list, optional (default: "Spectral")
        Custom color palette for plotting. Can be a string representing a seaborn palette or a list of colors.

    Returns
    -------
    matplotlib.pyplot.Figure
        A swarm plot showing cosine similarities for marker genes grouped by cell type.
    """
    # Access the gene expression matrix and mapping layer
    import pandas as pd
    import numpy as np
    import scanpy as sc
    from scipy.spatial.distance import cosine
    import matplotlib.pyplot as plt
    import seaborn as sns
    mapping_ori = adata.layers["mapping_ori"]  # Original mapping layer
    adata_x = adata.X  # Main gene expression matrix

    # Align gene symbols between the marker dataframe and the AnnData object
    genes = adata.var.index.tolist()  # List of genes in AnnData
    marker_genes = marker_df['gene_symbol'].tolist()  # List of marker genes
    common_genes = list(set(marker_genes).intersection(genes))  # Find common genes

    # Filter the marker DataFrame to only include the common genes
    marker_df = marker_df[marker_df['gene_symbol'].isin(common_genes)]

    # Initialize a list to store results for all cell types
    all_results = []

    # Group marker genes by cell type
    for cell_type, group in marker_df.groupby("cell_type"):
        print(f"Processing cell_type: {cell_type}")
        
        # Filter marker genes with p-value < 0.05
        filtered_group = group[group['pvalue'] < 0.05]
        
        # If more than 50 genes are available, select the top 50 based on the 'score' column
        top_genes = filtered_group.nlargest(50, "score")

        # Define a helper function to compute cosine similarity
        def compute_cosine_similarity(gene_symbol, mapping_ori, adata_x, var_names):
            """
            Compute cosine similarity for a given gene.

            Parameters:
            ----------
            gene_symbol : str
                The gene symbol to compute similarity for.
            mapping_ori : np.ndarray
                The original mapping layer array.
            adata_x : np.ndarray
                Gene expression matrix from AnnData.
            var_names : list
                List of gene names in AnnData.

            Returns:
            -------
            float
                Cosine similarity value or NaN if computation is not possible.
            """
            # If the gene is not found in AnnData, return NaN
            if gene_symbol not in var_names:
                return np.nan
            
            # Get the index of the gene in the AnnData object
            gene_idx = var_names.index(gene_symbol)
            
            # Extract expression vectors for the gene
            gene_expression_vector = adata_x[:, gene_idx].flatten()  # Expression vector in adata.X
            ori_vector = mapping_ori[:, gene_idx].flatten()  # Expression vector in mapping_ori

            # Compute cosine similarity if vectors are non-zero
            if np.any(gene_expression_vector) and np.any(ori_vector):
                similarity = 1 - cosine(gene_expression_vector, ori_vector)
            else:
                similarity = np.nan  # Return NaN for zero vectors
            
            return similarity

        # Compute cosine similarity for each gene in the top genes
        var_names = adata.var.index.tolist()  # Get gene names from AnnData
        cosine_similarities = [
            compute_cosine_similarity(gene_symbol, mapping_ori, adata.X, var_names)
            for gene_symbol in top_genes["gene_symbol"]
        ]
        
        # Add cosine similarity values to the DataFrame
        top_genes["cosine_similarity"] = cosine_similarities
        
        # Append the results for the current cell type
        all_results.append(top_genes)

    # Concatenate results from all cell types into a single DataFrame
    final_results = pd.concat(all_results, ignore_index=True)

    # Use seaborn's default palette if no custom palette is provided
  # Use seaborn's default color palette

    # Create a Swarm Plot for cosine similarity grouped by cell type
    plt.figure(figsize=(6, 4))  # Set the figure size to a smaller size (6x4 inches)
    sns.swarmplot(
        x="cell_type", 
        y="cosine_similarity", 
        data=final_results, 
        palette=custom_palette  # Use custom or default palette
    )

    # Add labels and customize the plot
    plt.xlabel("Cell Type", fontsize=12)
    plt.ylabel("Cosine Similarity", fontsize=12)
    plt.xticks(fontsize=10)  # Adjust font size for x-axis labels
    plt.tight_layout()  # Ensure the layout fits well

    # Return the matplotlib.pyplot object for further customization or saving
    return plt


"""
General utility functions
"""

# expose the API


from ._math import get_sum, pca,normalization_cpm,pear,rotate_matrix,convert_pvalue_to_asterisks
from ._math import normal_center_df as normal_center
from ._stimulation import bulk_simulation,bulk_simulation_case,st_simulation_case,st_simulation
from ._utils import compute_cluster_averages,compute_bulk_with_average_exp, data_dict_integration,filter_samples,compute_average_cosin,marker_integration,filter_gene,normalize_data
from ._read_data import check_paths
from ._evaluation import eval_fraction,eval_comparsion,eval_each_sample_mse


"""
Accessors for frequently used data within an :class:`~anndata.AnnData`.
"""

# expose the API
from ._plot import plot_batch_effect as batch_effect
from ._plot import plot_celltype_fraction_pie as celltype_fraction_pie
from ._plot import plot_celltype_fraction_heatmap as celltype_fraction_heatmap
from ._plot import plot_paired_violin as paired_violin
from ._plot import plot_reconstruction as reconstruction_corr
from ._plot import plot_he_cell_type as he_cell_type
from ._plot import plot_gene_similarity as gene_similarity


from setuptools import setup, find_packages

setup(
    name="CytoBulk",  
    version="0.1.14",  
    author="Xueying WANG",  
    author_email="your_email@example.com",  
    description="Integrating transcriptional data to decipher the tumor microenvironment with the graph frequency domain model",  
    long_description=open("README.md", encoding="utf-8").read(),  
    long_description_content_type="text/markdown", 
    url="https://github.com/kristaxying/CytoBulk",  
    packages=find_packages(exclude=["tests", "tests.*"]),
    package_data={
        "cytobulk": ["preprocessing/*.R"],  # 指定需要打包的 .R 文件
    },
    classifiers=[
        "Programming Language :: Python :: 3",  
        "License :: OSI Approved :: MIT License",  
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",  
    install_requires=[
        "anndata>=0.10.0",
        "cellpose>=3.0.10",
        "imageio",
        "matplotlib",
        "numpy>=1.23.0",
        "openslide-python",
        "ortools==9.3.10497",
        "pandas>=2.2.0",
        "Pillow",
        "POT==0.9.5",
        "rpy2>=3.5.0",
        "scanpy",
        "scikit-learn",
        "scipy",
        "seaborn",
        "torch>2.1.0",
        "torchvision",
        "tqdm",
        "requests",
        "openslide-python",
        "openslide-bin",
        "scikit-image",
        "igraph",
        "leidenalg"
    ],
)


import numpy as np
import pandas as pd
from .. import utils
from .. import get
from ._rpackage import find_marker_giotto
from ._filtering import qc_bulk_sc,qc_st_sc,high_variable_gene
from sklearn.metrics.pairwise import cosine_similarity
from os.path import exists
import json
import time




def _check_cell_type(sc_adata,cell_type_key="cell_type",cell_list=None):
    
    """
    check the cell type in single cell whether could be found in marker gene profile.

    Parameters
    ----------
    marker_gene : dataframe
        An :class:`~pandas.dataframe` marker gene in of each cell type.
        Each column is marker genes of one cell type. the first row should be name of cell types.
    sc_adata : ~anndata.AnnData
        An :class:`~anndata.AnnData` containing the sc rna expression to filter.
    cell_type_key : string.
        The `.obs` key where the annotation is stored in sc adata.

    Returns
    -------
    Returns the filtered marker gene list.
    
    """
    sc_cells = np.unique(sc_adata.obs[cell_type_key])
    if cell_list is not None:
        common_cell = np.intersect1d(np.array(cell_list), sc_cells, assume_unique=False, return_indices=False)
    else:
        common_cell = sc_cells
    '''
    if marker_data is not None and not isinstance(marker_data, pd.DataFrame):
        marker_cells = np.intersect1d(marker_data.columns, common_cell, assume_unique=False, return_indices=False)
    else:
        marker_cells = common_cell

    
    if len(marker_cells) != len(common_cell): # filter happened
        warnings.warn("In marker gene profile, could not find all the cell type of sc adata. ")
    
    if marker_data is not None:
        marker_gene = marker_data[marker_cells].to_dict('list')
    else:
        marker_gene = dict.fromkeys(common_cell)
    '''
    return common_cell




def _join_marker(bulk_adata,sc_adata,annotation_key,marker_list,out_dir='./',dataset_name='',python_path=None):

    """
    join the auto find marker and database marker together.

    Parameters
    ----------
    raw_sc_adata: anndata.AnnData
        An :class:`~anndata.AnnData` containing the raw expression.
    annotation_key: string
        The `.obs` key where the annotation is stored.: anndata.AnnData
    marker_dict:
        The marker gene dictionary from database.

    Returns
    -------
    Returns the marker gene dictionary from both database and auto-seeking.

    """
    
    if not exists(f'{out_dir}/{dataset_name}_marker.txt'):
        print("-------------------------Start finding cell type marker genes using Gitto--------------------------")
        start_t = time.perf_counter()
        marker_gene = find_marker_giotto(sc_adata,annotation_key,out_dir,dataset_name,python_path)
        print("-------------------------------Finish marker gene detection--------------------------------------")
        print(f'Time to finish marker gene detection: {round(time.perf_counter() - start_t, 2)} seconds')
    else:
        print(f'{out_dir}/{dataset_name}_marker.txt already exists, skipping find marker.')
    
    marker_gene = pd.read_csv(f"{out_dir}/{dataset_name}_marker.txt",sep='\t')
    common_gene = sc_adata.var_names.intersection(bulk_adata.var_names)

    
    
    return utils.marker_integration(common_gene,marker_gene,marker_list)



def _normalization_data(bulk_adata,sc_adata,scale_factors=None,trans_method='log',save=False,project='',out_dir='./'):
    """
    Normalization on bulk and sc adata.
        CPM = readsMappedToGene * 1/totalNumReads * 10^4
        totalNumReads       - total number of mapped reads of a sample
        readsMappedToGene   - number of reads mapped to a selected gene

    Parameters
    ----------
    bulk_data: dataframe
        An :class:`~anndata.AnnData` containing the expression to normalization.
    sc_adata: anndata.AnnData
        An :class:`~anndata.AnnData` containing the expression to normalization.
    scale_factors:
        The number of counts to normalize every observation to before computing profiles. If `None`, no normalization is performed. 
    trans_method:
        What transformation to apply to the expression before computing the profiles. 
        - `log`: log(x+1)
        - `None`: no transformation
        

    Returns
    -------
    Returns the normalizated bulk data (adata) and sc data (adata).

    """

    print("Start normalization")
    bulk_data = utils.normalization_cpm(bulk_adata,scale_factors=scale_factors,trans_method=trans_method)
    print("Finsh bulk normalization")
    sc_data = utils.normalization_cpm(sc_adata,scale_factors=scale_factors,trans_method=trans_method)
    print("Finsh sc normalization")
    
    if save:
        new_dir = utils.check_paths(out_dir+'/filtered')
        b_data = get.count_data(bulk_data)
        c_data = get.count_data(sc_data)
        b_data.to_csv(new_dir+f"/{project}_nor_bulk.txt",sep='\t')
        c_data.to_csv(new_dir+f"/{project}_nor_sc.txt",sep='\t')

    return sc_data,bulk_data

def _filter_cells(sc_adata, bulk_adata,num=10):
    cell_name = sc_adata.obs_names
    bulk_matrix = bulk_adata.X
    sc_matrix = sc_adata.X
    similarity_matrix = cosine_similarity(bulk_matrix,sc_matrix)
    top_k_indices = np.argsort(-similarity_matrix, axis=1)[:, :num]
    selected_indices = np.unique(top_k_indices.flatten())
    #if 10*bulk_adata.shape[0]<500:
    target_num=500
    #else:
        #target_num=10*bulk_adata.shape[0]
    while len(selected_indices)<target_num:
          num=int(num*1.25)
          top_k_indices = np.argsort(-similarity_matrix, axis=1)[:, :num]
          selected_indices = np.unique(top_k_indices.flatten())

    '''
    for i in range(pseudo_matrix.shape[0]):
        sample_cor = np.dot(pseudo_matrix[i,:],bulk_matrix.T)
        print(sample_cor)
        if sample_cor.max()>=cut_off:
            choose_index.append(sample_name[i])
    '''
    return cell_name[selected_indices]



def preprocessing(bulk_data,
                sc_adata,
                annotation_key,
                is_st:False,
                downsampling=True,
                marker_list=None,
                rename=None,
                dataset_name="",
                out_dir='.',
                different_source=True,
                cell_list=None,
                scale_factors=10000,
                trans_method="log",
                save = True,
                save_figure=True,
                n_sample_each_group=500,
                min_cells_each_group=100,
                cell_gap_each_group=100,
                group_number=5,
                filter_samples=False,
                python_path=None,
                **kwargs):
    """
    Preprocessing on bulk and sc adata, including following steps:\ 
        1. QC on bulk and sc adata.\ 
        2. Get common gene and common cell type.\  
        3. Get marker gene which is suitable for this dataset.\ 
        4. Normalization and scale.\ 
        5. Stimulation and batch effects.\ 
        6. NNLS to elimate gap between stimulated bulk and sc adata.\ 
        7. Transform gene expression value in input data.\ 

    Parameters
    ----------
    bulk_data: dataframe
        An :class:`~pandas.dataframe` containing the expression to normalization.
    sc_adata: anndata.AnnData
        An :class:`~anndata.AnnData` containing the expression to normalization.
    marker_data: 
        An :class:`~pandas.dataframe` which columns are cell types, rows are marker gene.
    annotation_key: string
        The `.obs` key where the single cell annotation is stored.: anndata.AnnData.
    project: string.
        The prefix of output file.
    out_dir: string, optional
        The path to store the output data.
    different_source: boolean, optional
        True for single cell and bulk data from the same sample, which means not executing batch effect.
        False for single cell and bulk data from the different samples, which means executing batch effect.
    cell_list: list, optional
        The list indicate the cell type names which need to take into consideration.
    scale_factors: int, optional
        The number of counts to normalize every observation to before computing profiles. If `None`, no normalization is performed. 
    trans_method: string, optional
        What transformation to apply to the expression before computing the profiles. 
        - "log": log(x+1)
        - `None`: no transformation
    save_figure: boolean, optional
        Whether save figures during preprocessing. eg. scatter plot of pca data.
    python_path: string, optional
        The path of python.exe for Giotto package.
    **kwargs: 
        Additional keyword arguments forwarded to
        :func:`~cytobulk.preprocessing.qc_bulk_sc`.
    

    Returns
    -------
    Returns the preprocessed bulk data (adata) containing the raw layer to store the raw data, 
    the batch_effected layer to fetch the expression after elimination of batch effect, and the transformed layer
    to fetch the expression value after linear transformation. Besides, adata.X stores the normalization data.
    adata.obsm['scale_factor'] saves the vector to normalize expression value.

    The stimualted bulk data (adata) using sc adata, and adata.X stores the normalization data. I also contains the
    normalization, batch_effected and transformed expression data. the cell_average_bulk key could get the bulk data
    stimulated with cell average expression. adata.obsm["sti_fraction"] stores the ground truth fraction.

    The preprocessed sc data (adata), which has .obsm["marker_gene"] to stores marker gene of valid cell types.

    """
    if  annotation_key not in sc_adata.obs:
        raise ValueError(f'The key {annotation_key!r} is not available in .obs!')
    # rename the cell type in sc_adata.obs
    if rename is not None:
        print("Start renaming cell type annotation")
        if not isinstance(rename,dict):
            raise ValueError(f"`A` can only be a dict but is a {rename.__class__}!")
        
        sc_adata.obs['curated_cell_type'] = sc_adata.obs.apply(lambda x: rename[x[annotation_key]] if x[annotation_key] in rename else "invalid", axis=1)
    else:
        sc_adata.obs.rename(columns={annotation_key: 'curated_cell_type'}, inplace=True)
        print("Finish renaming, the curated annotation could be found in sc_adata.obs['curated_cell_type']")
    annotation_key = 'curated_cell_type'
    print('================================================================================================================')
    print('------------------Start to check cell type annotation and quality control...------------------')
    sc_adata = sc_adata[sc_adata.obs[annotation_key]!="invalid",:]
    if is_st:
        sc_adata, bulk_adata, common_gene = qc_st_sc(bulk_data,sc_adata,out_dir=out_dir,save=save,dataset_name=dataset_name,**kwargs)
    else:
        sc_adata, bulk_adata, common_gene = qc_bulk_sc(bulk_data,sc_adata,out_dir=out_dir,save=save,dataset_name=dataset_name,**kwargs)

    common_cell = _check_cell_type(sc_adata,annotation_key,cell_list)
    print(common_cell)
    print('------------------Finish quality control------------------')

    if scale_factors is not None or trans_method is not None:
        sc_adata.layers["original"] = sc_adata.X
        bulk_adata.layers["original"] = bulk_adata.X
        sc_adata, bulk_adata = _normalization_data(bulk_adata,sc_adata,scale_factors,trans_method,
                                                save=False,project=dataset_name,out_dir=out_dir)

    # find informative genes.
    cell_type_counts = sc_adata.obs[annotation_key].value_counts()
    print("cell type after filtering:")
    print(cell_type_counts)
    if downsampling:
        print("start sampling")
        cell_types = sc_adata.obs[annotation_key].unique()
        sampled_indices = []
        for cell_type in cell_types:
            cell_type_indices = sc_adata.obs.index[sc_adata.obs[annotation_key] == cell_type].tolist()
            if 10*bulk_adata.shape[0]<3000:
                target_num=3000
            else:
                target_num=10*bulk_adata.shape[0]
            if len(cell_type_indices)<target_num:
                sampled_indices.extend(cell_type_indices)
            else:
                #sampled_indices.extend(_filter_cells(sc_adata[sc_adata.obs[annotation_key] == cell_type],bulk_adata))
                sampled_indices.extend(np.random.choice(cell_type_indices, target_num, replace=False))
                
        sc_adata = sc_adata[sampled_indices].copy()
        cell_type_counts = sc_adata.obs[annotation_key].value_counts()
        print("cell type after downsampling:")
        print(cell_type_counts)
        print("Finish")
    print('------------------Start to find hvg------------------')
    sc_adata = high_variable_gene(sc_adata,flavor='seurat')
    print(f'The number of valid genes of sc data is {sc_adata.shape[1]}')
    bulk_adata = high_variable_gene(bulk_adata,flavor='seurat')
    print(f'The number of valid genes of bulk data is {sc_adata.shape[1]}')
    print('------------------Finish------------------')
    common_gene = sc_adata.var.index.intersection(bulk_adata.var.index)
    # subset datasets
    print('Start to find vaild marker genes')
    if marker_list is not None:
        marker_list = common_gene.intersection(np.unique(np.array(marker_list)))
    else:
        marker_list = []
    sc_adata = sc_adata[:,common_gene]
    bulk_adata = bulk_adata[:,common_gene]
    '''
    if marker_list is not None:
        for i in db_marker.keys():
            flag=False
            if str(db_marker[i][0])!="nan":
                flag=True
            if not flag:
                db_gene=[]
            else:
                db_gene = np.array(db_marker[i])
            tmp = np.intersect1d(db_gene,common_gene)
            if len(db_marker) != len(tmp):
                db_marker[i] = tmp
    '''
    '''
    marker_dict,marker_gene,common_cell = _join_marker(bulk_adata,sc_adata,annotation_key,db_marker,common_cell,out_dir,dataset_name)
    '''
    marker_gene = _join_marker(bulk_adata,sc_adata,annotation_key,marker_list,out_dir,dataset_name)
    # debug
    print('Finish finding vaild marker genes')
    #marker_gene = np.intersect1d(np.array(marker_gene), sc_adata.var_names, assume_unique=False, return_indices=False)
    #bulk_adata = bulk_adata[:,marker_gene]
    bulk_adata = bulk_adata[:,marker_gene]
    sc_adata = sc_adata[sc_adata.obs[annotation_key].isin(common_cell),marker_gene]
    #sc_adata = sc_adata[sc_adata.obs[annotation_key].isin(common_cell),marker_gene]
    print(f'The number of valid single cell is {sc_adata.shape[0]}, valid sample is {bulk_adata.shape[0]},the number of valid genes is {sc_adata.shape[1]}')
    
    print('================================================================================================================')
    
    
    
    # training data simulation

    if is_st:
        if n_sample_each_group<800:
            n_sample_each_group=800
        pseudo_adata = utils.st_simulation(sc_adata, 
                bulk_adata,
                common_cell, 
                annotation_key = annotation_key,
                project=dataset_name, 
                out_dir=out_dir,
                n_sample_each_group=n_sample_each_group,
                min_cells_each_group=min_cells_each_group,
                cell_gap_each_group=cell_gap_each_group,
                group_number=group_number,
                average_ref=True,
                save=True,
                return_adata=True)
    else:
        if n_sample_each_group<400:
            n_sample_each_group=400
        pseudo_adata = utils.bulk_simulation(sc_adata,
                        bulk_adata,
                        common_cell, 
                        annotation_key = annotation_key,
                        project=dataset_name, 
                        out_dir=out_dir,
                        n_sample_each_group=n_sample_each_group,
                        min_cells_each_group=min_cells_each_group,
                        cell_gap_each_group=cell_gap_each_group,
                        group_number=group_number,
                        save=True,
                        return_adata=True)
    common_cell = pseudo_adata.obs.columns
    print("common_cell")
    print(common_cell)

    if save:
        out_dir = utils.check_paths(f'{out_dir}/filtered')
        pseudo_adata.write_h5ad(f"{out_dir}/pseudo_bulk_{dataset_name}.h5ad")
        sc_adata.write_h5ad(f"{out_dir}/sc_data_{dataset_name}.h5ad")
        bulk_adata.write_h5ad(f"{out_dir}/bulk_data_{dataset_name}.h5ad")
        with open(f"{out_dir}/cells_{dataset_name}.json", "w") as outfile: 
            json.dump(list(common_cell), outfile)

    return sc_adata, pseudo_adata, bulk_adata, common_cell, annotation_key

    
